from __future__ import annotations

from django_bird.compiler import Compiler


def test_compiler():
    component_template = """
<style>
    body { font-family: Arial, sans-serif; }
</style>

<h1>Welcome to my page</h1>
{{ user.name }}

<script>
    console.log('Hello, world!');
</script>

<p>This is some more content.</p>

<style>
    .highlight { color: red; }
</style>
"""
    compiler = Compiler()

    tokens = compiler.tokenize(component_template)
    ast = compiler.parse(tokens)
    transformed = compiler.transform(ast)
    compiled = compiler.compile(component_template)

    assert tokens
    assert ast
    assert transformed
    assert compiled



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from cachetools import LRUCache
from django.template.backends.django import Template
from django.template.loader import select_template

from .templates import get_template_names


@dataclass(frozen=True, slots=True)
class Component:
    name: str
    template: Template

    def render(self, context: dict[str, Any]):
        return self.template.render(context)

    @property
    def nodelist(self):
        return self.template.template.nodelist

    @classmethod
    def from_name(cls, name: str):
        template_names = get_template_names(name)
        template = select_template(template_names)
        return cls(name=name, template=template)


class Registry:
    def __init__(self, maxsize: int = 100):
        self._cache: LRUCache[str, Component] = LRUCache(maxsize=maxsize)

    def get_component(self, name: str) -> Component:
        try:
            return self._cache[name]
        except KeyError:
            component = Component.from_name(name)
            self._cache[name] = component
            return component

    def clear(self) -> None:
        """Clear the component cache. Mainly useful for testing."""
        self._cache.clear()


components = Registry()



from __future__ import annotations

import logging
import re
from pathlib import Path

import pytest
from django.conf import settings
from django.test import override_settings

from .settings import DEFAULT_SETTINGS

pytest_plugins = []


def pytest_configure(config):
    logging.disable(logging.CRITICAL)

    settings.configure(**DEFAULT_SETTINGS, **TEST_SETTINGS)


TEST_SETTINGS = {
    "INSTALLED_APPS": [
        "django_bird",
    ],
    "TEMPLATES": [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": [
                Path(__file__).parent / "templates",
            ],
            "OPTIONS": {
                "builtins": [
                    "django.template.defaulttags",
                    "django_bird.templatetags.django_bird",
                ],
                "loaders": [
                    "django_bird.loader.BirdLoader",
                    "django.template.loaders.filesystem.Loader",
                    "django.template.loaders.app_directories.Loader",
                ],
            },
        }
    ],
}


@pytest.fixture
def base_dir(tmp_path):
    base_dir = tmp_path / "templates"
    base_dir.mkdir()
    return base_dir


@pytest.fixture(autouse=True)
def override_templates_settings(base_dir):
    with override_settings(
        TEMPLATES=[
            settings.TEMPLATES[0]
            | {
                "DIRS": [
                    *settings.TEMPLATES[0]["DIRS"],
                    base_dir,
                ]
            }
        ]
    ):
        yield


@pytest.fixture
def create_bird_dir(base_dir):
    def func(name):
        bird_template_dir = base_dir / name
        bird_template_dir.mkdir(exist_ok=True)
        return bird_template_dir

    return func


@pytest.fixture
def create_bird_template(create_bird_dir):
    def func(name, content, sub_dir=None, bird_dir_name="bird"):
        bird_dir = create_bird_dir(bird_dir_name)
        if sub_dir is not None:
            dir = bird_dir / sub_dir
            dir.mkdir()
        else:
            dir = bird_dir
        template = dir / f"{name}.html"
        template.write_text(content)
        return template

    return func


@pytest.fixture
def normalize_whitespace():
    def func(text):
        # this makes writing tests much easier, as it gets rid of any
        # existing whitespace that may be present in the template file

        # multiple whitespace characters
        text = re.sub(r"\s+", " ", text)

        # after opening tag, including when there are attributes
        text = re.sub(r"<(\w+)(\s+[^>]*)?\s*>", r"<\1\2>", text)

        # before closing tag
        text = re.sub(r"\s+>", ">", text)

        # after opening tag and before closing tag
        text = re.sub(r">\s+<", "><", text)

        # immediately after opening tag (including attributes) or before closing tag
        text = re.sub(r"(<\w+(?:\s+[^>]*)?>)\s+|\s+(<\/\w+>)", r"\1\2", text)

        return text.strip()

    return func


@pytest.fixture(autouse=True)
def clear_registry():
    from django_bird.components import components

    components.clear()
    yield
    components.clear()



from __future__ import annotations

import hashlib

from django.core.cache import cache
from django.template.base import Origin
from django.template.engine import Engine
from django.template.loaders.filesystem import Loader as FileSystemLoader

from ._typing import override
from .compiler import BIRD_PATTERN
from .compiler import Compiler


class BirdLoader(FileSystemLoader):
    def __init__(self, engine: Engine):
        super().__init__(engine)
        self.compiler = Compiler()

    @override
    def get_contents(self, origin: Origin) -> str:
        contents = super().get_contents(origin)
        if not BIRD_PATTERN.search(contents):
            return contents
        cache_key = f"bird_component_{hashlib.md5(contents.encode()).hexdigest()}"
        compiled = cache.get(cache_key)  # pyright: ignore[reportAny]
        if compiled is None:
            compiled = self.compiler.compile(contents)
            cache.set(cache_key, compiled, timeout=None)
        return compiled



from __future__ import annotations

from pathlib import Path

import pytest
from django.utils.safestring import SafeString

from django_bird.staticfiles import Asset
from django_bird.staticfiles import AssetType
from django_bird.staticfiles import registry


class TestAsset:
    def test_hash(self):
        asset1 = Asset(Path("static.css"), AssetType.CSS)
        asset2 = Asset(Path("static.css"), AssetType.CSS)

        assert asset1 == asset2
        assert hash(asset1) == hash(asset2)

        assets = {asset1, asset2, Asset(Path("other.css"), AssetType.CSS)}

        assert len(assets) == 2

    def test_exists(self, tmp_path: Path):
        css_file = tmp_path / "test.css"
        css_file.touch()

        asset = Asset(css_file, AssetType.CSS)

        assert asset.exists() is True

    def test_exists_nonexistent(self):
        missing_asset = Asset(Path("missing.css"), AssetType.CSS)
        assert missing_asset.exists() is False

    @pytest.mark.parametrize(
        "path,expected",
        [
            (Path("static.css"), Asset(Path("static.css"), AssetType.CSS)),
            (Path("static.js"), Asset(Path("static.js"), AssetType.JS)),
            (
                Path("nested/path/style.css"),
                Asset(Path("nested/path/style.css"), AssetType.CSS),
            ),
            (
                Path("./relative/script.js"),
                Asset(Path("./relative/script.js"), AssetType.JS),
            ),
            (Path("UPPERCASE.CSS"), Asset(Path("UPPERCASE.CSS"), AssetType.CSS)),
            (Path("mixed.Js"), Asset(Path("mixed.Js"), AssetType.JS)),
        ],
    )
    def test_from_path(self, path, expected):
        assert Asset.from_path(path) == expected

    def test_from_path_invalid(self):
        with pytest.raises(ValueError):
            Asset.from_path(Path("static.html"))


class TestRegistry:
    @pytest.fixture
    def registry(self):
        registry.clear()
        yield registry
        registry.clear()

    def test_register_asset(self, registry, tmp_path):
        css_file = tmp_path / "test.css"
        css_file.touch()
        asset = Asset(css_file, AssetType.CSS)

        registry.register(asset)

        assert asset in registry.assets

    def test_register_path(self, registry, tmp_path):
        css_file = tmp_path / "test.css"
        css_file.touch()

        registry.register(css_file)

        assert len(registry.assets) == 1
        assert next(iter(registry.assets)).path == css_file

    def test_register_missing_file(self, registry, tmp_path):
        missing_file = tmp_path / "missing.css"

        with pytest.raises(FileNotFoundError):
            registry.register(missing_file)

    def test_clear(self, registry, tmp_path):
        css_file = tmp_path / "test.css"
        css_file.touch()
        js_file = tmp_path / "test.js"
        js_file.touch()
        registry.register(Asset(css_file, AssetType.CSS))
        registry.register(Asset(js_file, AssetType.JS))

        assert len(registry.assets) == 2

        registry.clear()

        assert len(registry.assets) == 0

    @pytest.mark.parametrize("asset_type", [AssetType.CSS, AssetType.JS])
    def test_get_assets(self, asset_type, registry):
        css_asset = Asset(Path("test.css"), AssetType.CSS)
        js_asset = Asset(Path("test.js"), AssetType.JS)

        registry.assets = {css_asset, js_asset}

        assets = registry.get_assets(asset_type)

        assert len(assets) == 1
        assert all(asset.type == asset_type for asset in assets)

    @pytest.mark.parametrize(
        "assets,sort_key,expected",
        [
            (
                [
                    Asset(Path("test/a.css"), AssetType.CSS),
                    Asset(Path("test/b.css"), AssetType.CSS),
                ],
                None,
                ["test/a.css", "test/b.css"],
            ),
            (
                [
                    Asset(Path("test/b.css"), AssetType.CSS),
                    Asset(Path("other/a.css"), AssetType.CSS),
                ],
                lambda a: a.path.name,
                ["other/a.css", "test/b.css"],
            ),
            ([], None, []),
            (
                [Asset(Path("test.css"), AssetType.CSS)],
                None,
                ["test.css"],
            ),
        ],
    )
    def test_sort_assets(self, assets, sort_key, expected, registry):
        kwargs = {}
        if sort_key is not None:
            kwargs["key"] = sort_key

        sorted_assets = registry.sort_assets(assets, **kwargs)

        assert [str(a.path) for a in sorted_assets] == expected

    @pytest.mark.parametrize(
        "asset_type,expected",
        [
            (AssetType.CSS, '<link rel="stylesheet" href="{}" type="text/css">'),
            (AssetType.JS, '<script src="{}" type="text/javascript">'),
        ],
    )
    def test_get_format_string(self, asset_type, expected, registry):
        assert registry.get_format_string(asset_type) == expected

    @pytest.mark.parametrize(
        "assets,asset_type,expected",
        [
            (set(), AssetType.CSS, ""),
            (
                {Asset(Path("test.css"), AssetType.CSS)},
                AssetType.CSS,
                ('rel="stylesheet"', 'href="test.css"'),
            ),
            (
                {Asset(Path("test.js"), AssetType.JS)},
                AssetType.JS,
                ('<script src="test.js"',),
            ),
            (
                {
                    Asset(Path("a.css"), AssetType.CSS),
                    Asset(Path("b.css"), AssetType.CSS),
                },
                AssetType.CSS,
                ('stylesheet" href="a.css"', 'stylesheet" href="b.css"'),
            ),
        ],
    )
    def test_render(self, assets, asset_type, expected, registry):
        registry.assets = assets

        rendered = registry.render(asset_type)

        assert isinstance(rendered, SafeString)
        for content in expected:
            assert content in rendered



from __future__ import annotations

import pytest
from django.template import Context
from django.template import Template
from django.template.exceptions import TemplateSyntaxError

from django_bird.templatetags.tags.slot import parse_slot_name


@pytest.mark.parametrize(
    "bits,expected",
    [
        (["slot"], "default"),
        (["slot", "foo"], "foo"),
        (["slot", "'foo'"], "foo"),
        (["slot", '"foo"'], "foo"),
        (["slot", 'name="foo"'], "foo"),
        (["slot", "name='foo'"], "foo"),
    ],
)
def test_parse_slot_name(bits, expected):
    assert parse_slot_name(bits) == expected


def test_parse_slot_name_no_args():
    with pytest.raises(TemplateSyntaxError):
        assert parse_slot_name([])


class TestTemplateTag:
    @pytest.mark.parametrize(
        "template,context,expected",
        [
            ("{{ slot }}", {"slot": "test"}, "test"),
            ("{% bird:slot %}{% endbird:slot %}", {"slot": "test"}, "test"),
            ("{% bird:slot default %}{% endbird:slot %}", {"slot": "test"}, "test"),
            ("{% bird:slot 'default' %}{% endbird:slot %}", {"slot": "test"}, "test"),
            ('{% bird:slot "default" %}{% endbird:slot %}', {"slot": "test"}, "test"),
            (
                "{% bird:slot name='default' %}{% endbird:slot %}",
                {"slot": "test"},
                "test",
            ),
            (
                '{% bird:slot name="default" %}{% endbird:slot %}',
                {"slot": "test"},
                "test",
            ),
            (
                "{% bird:slot name='not-default' %}{% endbird:slot %}",
                {"slot": "test"},
                "",
            ),
            (
                "{% bird:slot outer %}Outer {% bird:slot inner %}Inner{% endbird:slot %} Content{% endbird:slot %}",
                {"slots": {"outer": "Replaced Content"}},
                "Replaced Content",
            ),
            (
                "{% bird:slot outer %}Outer {% bird:slot inner %}Inner{% endbird:slot %} Content{% endbird:slot %}",
                {"slots": {"inner": "Replaced Content"}},
                "Outer Replaced Content Content",
            ),
            (
                "{% bird:slot outer %}Outer {% bird:slot inner %}Inner Default{% endbird:slot %} Content{% endbird:slot %}",
                {
                    "slots": {
                        "outer": "Replaced {% bird:slot inner %}{% endbird:slot %} Outer",
                        "inner": "Replaced Inner",
                    },
                },
                "Replaced Replaced Inner Outer",
            ),
            (
                "{{ slot }}",
                {
                    "slot": "Replaced {% bird:slot inner %}{% endbird:slot %} Outer",
                    "slots": {
                        "inner": "Replaced Inner",
                    },
                },
                "Replaced Replaced Inner Outer",
            ),
            (
                "{% bird:slot %}{% endbird:slot %}",
                {},
                "",
            ),
            (
                "{% bird:slot %}Default content{% endbird:slot %}",
                {},
                "Default content",
            ),
            (
                "{% bird:slot 'mal formed' %}{% endbird:slot %}",
                {"slots": {"mal formed": "content"}},
                "content",
            ),
            (
                "{% bird:slot CaseSensitive %}{% endbird:slot %}",
                {"slots": {"CaseSensitive": "Upper", "casesensitive": "Lower"}},
                "Upper",
            ),
            (
                "{% bird:slot %}{% endbird:slot %}",
                {"slots": {"default": 42}},
                "42",
            ),
            (
                "{% bird:slot %}{% endbird:slot %}",
                {"slots": {"default": "<b>Bold</b>"}},
                "<b>Bold</b>",
            ),
            (
                "{% bird:slot unicode_slot %}{% endbird:slot %}",
                {"slots": {"unicode_slot": "こんにちは"}},
                "こんにちは",
            ),
        ],
    )
    def test_rendering(self, template, context, expected, create_bird_template):
        create_bird_template("test", template)
        t = Template(f"{{% bird test %}}{template}{{% endbird %}}")
        rendered = t.render(context=Context(context))
        assert rendered == expected

    def test_too_many_args(self):
        with pytest.raises(TemplateSyntaxError):
            Template("{% bird:slot too many args %}{% endbird:slot %}")



from __future__ import annotations

from .conf import app_settings


def get_template_names(name: str) -> list[str]:
    """
    Generate a list of potential template names for a component.

    The function searches for templates in the following order (from most specific to most general):

    1. In a subdirectory named after the component, using the component name
    2. In the same subdirectory, using a fallback 'index.html'
    3. In parent directory for nested components
    4. In the base component directory, using the full component name

    The order of names is important as it determines the template resolution priority.
    This order allows for both direct matches and hierarchical component structures,
    with more specific paths taking precedence over more general ones.

    This order allows for:
    - Single file components
    - Multi-part components
    - Specific named files within component directories
    - Fallback default files for components

    For example:
    - For an "input" component, the ordering would be:
        1. `{component_dir}/input/input.html`
        2. `{component_dir}/input/index.html`
        3. `{component_dir}/input.html`
    - For an "input.label" component:
        1. `{component_dir}/input/label/label.html`
        2. `{component_dir}/input/label/index.html`
        3. `{component_dir}/input/label.html`
        4. `{component_dir}/input.label.html`

    Returns:
        list[str]: A list of potential template names in resolution order.
    """
    template_names = []
    component_dirs = list(dict.fromkeys([*app_settings.COMPONENT_DIRS, "bird"]))

    name_parts = name.split(".")
    path_name = "/".join(name_parts)

    for component_dir in component_dirs:
        potential_names = [
            f"{component_dir}/{path_name}/{name_parts[-1]}.html",
            f"{component_dir}/{path_name}/index.html",
            f"{component_dir}/{path_name}.html",
            f"{component_dir}/{name}.html",
        ]
        template_names.extend(potential_names)

    return list(dict.fromkeys(template_names))



from __future__ import annotations

from django.apps import AppConfig

from ._typing import override


class DjangoBirdAppConfig(AppConfig):
    label = "django_bird"
    name = "django_bird"
    verbose_name = "Bird"

    @override
    def ready(self):
        from django_bird.conf import app_settings

        app_settings.autoconfigure()



# pyright: reportAny=false
from __future__ import annotations

from typing import Any

from django import template
from django.template.base import NodeList
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context

from django_bird._typing import TagBits
from django_bird._typing import override
from django_bird.components import Component
from django_bird.components import components
from django_bird.params import Params
from django_bird.slots import DEFAULT_SLOT
from django_bird.slots import Slots

TAG = "bird"
END_TAG = "endbird"


def do_bird(parser: Parser, token: Token) -> BirdNode:
    bits = token.split_contents()
    name = parse_bird_name(bits)
    params = Params.from_bits(bits[2:])
    nodelist = parse_nodelist(bits, parser)
    return BirdNode(name, params, nodelist)


def parse_bird_name(bits: TagBits) -> str:
    if len(bits) == 1:
        msg = f"{TAG} tag requires at least one argument"
        raise template.TemplateSyntaxError(msg)

    # {% bird name %}
    # {% bird 'name' %}
    # {% bird "name" %}
    return bits[1].strip("'\"")


def parse_nodelist(bits: TagBits, parser: Parser) -> NodeList | None:
    # self-closing tag
    # {% bird name / %}
    if len(bits) > 0 and bits[-1] == "/":
        nodelist = None
    else:
        nodelist = parser.parse((END_TAG,))
        parser.delete_first_token()
    return nodelist


class BirdNode(template.Node):
    def __init__(self, name: str, params: Params, nodelist: NodeList | None) -> None:
        self.name = name
        self.params = params
        self.nodelist = nodelist

    @override
    def render(self, context: Context) -> str:
        component_name = self.get_component_name(context)
        component = components.get_component(component_name)
        component_context = self.get_component_context_data(component, context)
        return component.render(component_context)

    def get_component_name(self, context: Context) -> str:
        try:
            name = template.Variable(self.name).resolve(context)
        except template.VariableDoesNotExist:
            name = self.name
        return name

    def get_component_context_data(
        self, component: Component, context: Context
    ) -> dict[str, Any]:
        props = self.params.render_props(component.nodelist, context)
        attrs = self.params.render_attrs(context)
        slots = Slots.collect(self.nodelist, context).render()
        default_slot = slots.get(DEFAULT_SLOT) or context.get("slot")
        return {
            "attrs": attrs,
            "props": props,
            "slot": default_slot,
            "slots": slots,
        }



from __future__ import annotations

import pytest
from django.template.loader import get_template


@pytest.mark.parametrize(
    "template_name",
    (
        "with_bird.html",
        "without_bird.html",
    ),
)
def test_render_template(template_name):
    context = {
        "title": "Test Title",
        "content": "Test Content",
    }

    template = get_template(template_name)
    rendered = template.render(context)

    assert rendered



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field

from django import template
from django.template.base import NodeList
from django.template.context import Context
from django.utils.safestring import SafeString
from django.utils.safestring import mark_safe

from ._typing import TagBits


@dataclass
class Param:
    name: str
    value: str | bool | None

    def render(self, context: Context) -> str:
        if self.value is None or (
            isinstance(self.value, str) and self.value == "False"
        ):
            return ""
        if (isinstance(self.value, bool) and self.value) or (
            isinstance(self.value, str) and self.value == "True"
        ):
            return self.name
        try:
            value = template.Variable(str(self.value)).resolve(context)
        except template.VariableDoesNotExist:
            value = self.value
        return f'{self.name}="{value}"'

    @classmethod
    def from_bit(cls, bit: str):
        value: str | bool
        if "=" in bit:
            name, value = bit.split("=", 1)
            value = value.strip("'\"")
        else:
            name, value = bit, True
        return cls(name, value)


@dataclass
class Params:
    attrs: list[Param] = field(default_factory=list)
    props: list[Param] = field(default_factory=list)

    def render_props(self, nodelist: NodeList | None, context: Context):
        from django_bird.templatetags.tags.prop import PropNode

        if nodelist is None:
            return

        attrs_to_remove = set()

        for node in nodelist:
            if not isinstance(node, PropNode):
                continue

            value: str | bool | None = node.default

            for idx, attr in enumerate(self.attrs):
                if node.name == attr.name:
                    if attr.value:
                        value = attr.value
                    attrs_to_remove.add(idx)

            self.props.append(Param(name=node.name, value=value))

        for idx in sorted(attrs_to_remove, reverse=True):
            self.attrs.pop(idx)

        return {prop.name: prop.value for prop in self.props}

    def render_attrs(self, context: Context) -> SafeString:
        rendered = " ".join(attr.render(context) for attr in self.attrs)
        return mark_safe(rendered)

    @classmethod
    def from_bits(cls, bits: TagBits):
        params: list[Param] = []
        for bit in bits:
            param = Param.from_bit(bit)
            params.append(param)
        return cls(attrs=params)



from __future__ import annotations

from unittest.mock import patch

import pytest
from django.template import Context
from django.template import Library
from django.template import Template
from django.template.base import Parser
from django.template.base import Token
from django.template.base import TokenType
from django.template.exceptions import TemplateSyntaxError

from django_bird.params import Param
from django_bird.params import Params
from django_bird.templatetags.tags.bird import END_TAG
from django_bird.templatetags.tags.bird import TAG
from django_bird.templatetags.tags.bird import BirdNode
from django_bird.templatetags.tags.bird import do_bird


class TestTemplateTag:
    @pytest.mark.parametrize(
        "name,expected",
        [
            ("button", "button"),
            ("'button'", "button"),
            ('"button"', "button"),
            ("button.label", "button.label"),
        ],
    )
    def test_node_name(self, name, expected):
        token = Token(TokenType.BLOCK, f"{TAG} {name}")
        parser = Parser(
            [Token(TokenType.BLOCK, END_TAG)],
        )
        node = do_bird(parser, token)
        assert node.name == expected

    def test_missing_argument(self):
        token = Token(TokenType.BLOCK, TAG)
        parser = Parser(
            [Token(TokenType.BLOCK, END_TAG)],
        )
        with pytest.raises(TemplateSyntaxError):
            do_bird(parser, token)

    @pytest.mark.parametrize(
        "params,expected_params",
        [
            ("class='btn'", Params(attrs=[Param(name="class", value="btn")])),
            (
                "class='btn' id='my-btn'",
                Params(
                    attrs=[
                        Param(name="class", value="btn"),
                        Param(name="id", value="my-btn"),
                    ]
                ),
            ),
            ("disabled", Params(attrs=[Param(name="disabled", value=True)])),
        ],
    )
    def test_node_params(self, params, expected_params):
        token = Token(TokenType.BLOCK, f"{TAG} button {params}")
        parser = Parser(
            [Token(TokenType.BLOCK, END_TAG)],
        )
        node = do_bird(parser, token)
        assert node.params == expected_params

    @pytest.mark.parametrize(
        "component,template,context,expected",
        [
            (
                "<button>Click me</button>",
                "{% bird button %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button>Click me</button>",
                "{% bird 'button' %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button>Click me</button>",
                '{% bird "button" %}Click me{% endbird %}',
                {},
                "<button>Click me</button>",
            ),
            (
                "<button>Click me</button>",
                "{% with dynamic_name='button' %}{% bird dynamic_name %}Click me{% endbird %}{% endwith %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button>Click me</button>",
                "{% bird dynamic-name %}Click me{% endbird %}",
                {"dynamic-name": "button"},
                "<button>Click me</button>",
            ),
        ],
    )
    def test_rendered_name(
        self,
        component,
        template,
        context,
        expected,
        create_bird_template,
        normalize_whitespace,
    ):
        create_bird_template("button", component)
        t = Template(template)
        rendered = t.render(context=Context(context))
        assert normalize_whitespace(rendered) == expected

    @pytest.mark.parametrize(
        "component,template,context,expected",
        [
            (
                "<button {{ attrs }}>Click me</button>",
                "{% bird button class='btn' %}Click me{% endbird %}",
                {},
                '<button class="btn">Click me</button>',
            ),
            (
                "<button {{ attrs }}>Click me</button>",
                "{% bird button disabled %}Click me{% endbird %}",
                {},
                "<button disabled>Click me</button>",
            ),
            (
                "<button {{ attrs }}>Click me</button>",
                "{% bird button disabled=True %}Click me{% endbird %}",
                {},
                "<button disabled>Click me</button>",
            ),
            (
                "<button {{ attrs }}>Click me</button>",
                "{% bird button disabled=False %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button {{ attrs }}>Click me</button>",
                "{% bird button class='btn' id='my-btn' disabled %}Click me{% endbird %}",
                {},
                '<button class="btn" id="my-btn" disabled>Click me</button>',
            ),
            (
                "<button {{ attrs }}>{{ slot }}</button>",
                "{% bird button %}{{ slot }}{% endbird %}",
                {"slot": "Click me"},
                "<button>Click me</button>",
            ),
        ],
    )
    def test_rendered_attrs(
        self,
        component,
        template,
        context,
        expected,
        create_bird_template,
        normalize_whitespace,
    ):
        create_bird_template("button", component)
        t = Template(template)
        rendered = t.render(context=Context(context))
        assert normalize_whitespace(rendered) == expected

    def test_self_closing_tag(self, create_bird_template):
        create_bird_template(name="image", content="<img src='image' />")
        template = Template("{% bird image src='image' / %}")
        rendered = template.render(context=Context({}))
        assert rendered == "<img src='image' />"

    @pytest.mark.parametrize(
        "component,template,context,expected",
        [
            (
                "<button>{{ slot }}</button>",
                "{% bird button %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button>{% bird:slot %}{% endbird:slot %}</button>",
                "{% bird button %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button>{% bird:slot default %}{% endbird:slot %}</button>",
                "{% bird button %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button><span>{% bird:slot leading-icon %}{% endbird:slot %}</span>{{ slot }}</button>",
                "{% bird button %}{% bird:slot leading-icon %}Icon here{% endbird:slot %}Click me{% endbird %}",
                {},
                "<button><span>Icon here</span>Click me</button>",
            ),
        ],
    )
    def test_with_slots(
        self,
        component,
        template,
        context,
        expected,
        create_bird_template,
        normalize_whitespace,
    ):
        create_bird_template("button", component)
        t = Template(template)
        rendered = t.render(context=Context(context))
        assert normalize_whitespace(rendered) == expected

    @pytest.mark.parametrize(
        "component,template,context,expected",
        [
            (
                "<button>{{ slot }}</button>",
                "{% load custom_filters %}{% bird button %}{{ text|make_fancy }}{% endbird %}",
                {"text": "click me"},
                "<button>✨click me✨</button>",
            ),
            (
                "<button>{% bird:slot label %}{% endbird:slot %}</button>",
                "{% load custom_filters %}{% bird button %}{% bird:slot label %}{{ text|make_fancy }}{% endbird:slot %}{% endbird %}",
                {"text": "click me"},
                "<button>✨click me✨</button>",
            ),
            (
                "<button><span>{% bird:slot icon %}{% endbird:slot %}</span>{{ slot }}</button>",
                "{% load custom_filters %}{% bird button %}{% bird:slot icon %}{{ icon_text|make_fancy }}{% endbird:slot %}{{ text|make_fancy }}{% endbird %}",
                {"text": "click me", "icon_text": "icon"},
                "<button><span>✨icon✨</span>✨click me✨</button>",
            ),
        ],
    )
    def test_slots_with_outside_templatetag(
        self,
        component,
        template,
        context,
        expected,
        create_bird_template,
        normalize_whitespace,
    ):
        register = Library()

        @register.filter
        def make_fancy(value):
            return f"✨{value}✨"

        def get_template_libraries(self, libraries):
            return {"custom_filters": register}

        with patch(
            "django.template.engine.Engine.get_template_libraries",
            get_template_libraries,
        ):
            create_bird_template("button", component)
            t = Template(template)
            rendered = t.render(context=Context(context))
            assert normalize_whitespace(rendered) == expected

    @pytest.mark.parametrize(
        "component,template,context,expected",
        [
            (
                "{% bird:prop id %}<button id='{{ props.id }}'>{{ slot }}</button>",
                "{% bird button id='foo' %}Click me{% endbird %}",
                {},
                "<button id='foo'>Click me</button>",
            ),
            (
                "{% bird:prop id='default' %}<button id='{{ props.id }}'>{{ slot }}</button>",
                "{% bird button %}Click me{% endbird %}",
                {},
                "<button id='default'>Click me</button>",
            ),
            (
                "{% bird:prop id='default' %}<button id='{{ props.id }}'>{{ slot }}</button>",
                "{% bird button id='foo' %}Click me{% endbird %}",
                {},
                "<button id='foo'>Click me</button>",
            ),
            (
                '{% bird:prop id="default" %}{% bird:prop class="btn" %}<button id="{{ props.id }}" class="{{ props.class }}" {{ attrs }}>{{ slot }}</button>',
                "{% bird button data-test='value' %}Click me{% endbird %}",
                {},
                '<button id="default" class="btn" data-test="value">Click me</button>',
            ),
            (
                '{% bird:prop class="btn" %}{% bird:prop data_attr="button" %}<button class="{{ props.class }}" data-attr="{{ props.data_attr }}" {{ attrs }}>{{ slot }}</button>',
                '{% bird button class="btn-primary" disabled=True %}Click me{% endbird %}',
                {},
                '<button class="btn-primary" data-attr="button" disabled>Click me</button>',
            ),
        ],
    )
    def test_with_props(
        self,
        component,
        template,
        context,
        expected,
        create_bird_template,
        normalize_whitespace,
    ):
        create_bird_template("button", component)
        t = Template(template)
        rendered = t.render(context=Context(context))
        assert normalize_whitespace(rendered) == expected


@pytest.mark.xfail(reason="Feature not implemented yet")
class TestTemplateTagFutureFeatures:
    def test_dynamic_attrs(self, create_bird_template):
        create_bird_template(
            name="button", content="<button class=btn_class>Click me</button>"
        )
        template = Template("{% bird button class=btn_class %}Click me{% endbird %}")
        rendered = template.render(context=Context({"btn_class": "primary"}))
        assert rendered == "<button class='primary'>Click me</button>"


class TestNode:
    @pytest.mark.parametrize(
        "name,context,expected",
        [
            (
                "button",
                {},
                "button",
            ),
            (
                "dynamic-name",
                {"dynamic-name": "button"},
                "button",
            ),
        ],
    )
    def test_get_component_name(self, name, context, expected, create_bird_template):
        create_bird_template(name=name, content="<button>Click me</button>")
        node = BirdNode(name=name, params=Params([]), nodelist=None)

        component_name = node.get_component_name(context=Context(context))

        assert component_name == expected



from __future__ import annotations

import re

BIRD_PATTERN = re.compile(
    r"<bird:(\w+)([^>]*)(?:/>|>(.*?)</bird:\1>)", re.DOTALL | re.MULTILINE
)


class Compiler:
    # placeholder for now, implementation to come
    def compile(self, input_string: str) -> str:
        tokens = self.tokenize(input_string)
        ast = self.parse(tokens)
        return self.transform(ast)

    def tokenize(self, input_string: str) -> str:
        return input_string

    def parse(self, tokens: str) -> str:
        return tokens

    def transform(self, ast: str) -> str:
        return ast



from __future__ import annotations

import pytest
from django.template.backends.django import Template
from django.template.exceptions import TemplateDoesNotExist

from django_bird.components import Component
from django_bird.components import Registry


class TestComponent:
    def test_from_name(self, create_bird_template):
        create_bird_template("button", "<button>Click me</button>")

        comp = Component.from_name("button")

        assert comp.name == "button"
        assert isinstance(comp.template, Template)
        assert comp.render({}) == "<button>Click me</button>"


class TestRegistry:
    @pytest.fixture
    def registry(self):
        return Registry(maxsize=2)

    def test_get_component_caches(self, registry, create_bird_template):
        create_bird_template(name="button", content="<button>Click me</button>")

        component1 = registry.get_component("button")
        component2 = registry.get_component("button")

        assert component1 is component2

    def test_lru_cache_behavior(self, registry, create_bird_template):
        create_bird_template(name="button1", content="1")
        create_bird_template(name="button2", content="2")
        create_bird_template(name="button3", content="3")

        button1 = registry.get_component("button1")
        button2 = registry.get_component("button2")
        button3 = registry.get_component("button3")

        new_button1 = registry.get_component("button1")
        assert new_button1 is not button1

        cached_button2 = registry.get_component("button2")
        assert cached_button2.name == button2.name
        assert cached_button2.render({}) == button2.render({})

        cached_button3 = registry.get_component("button3")
        assert cached_button3.name == button3.name
        assert cached_button3.render({}) == button3.render({})

    def test_component_not_found(self, registry):
        with pytest.raises(TemplateDoesNotExist):
            registry.get_component("nonexistent")



from __future__ import annotations

import pytest

from django_bird.params import Param
from django_bird.params import Params
from django_bird.templatetags.tags.prop import PropNode


class TestParam:
    @pytest.mark.parametrize(
        "param,context,expected",
        [
            (Param(name="class", value="btn"), {}, 'class="btn"'),
            (Param(name="class", value="btn"), {"btn": "blue"}, 'class="blue"'),
            (Param(name="disabled", value=True), {}, "disabled"),
            (Param(name="disabled", value=None), {}, ""),
        ],
    )
    def test_render(self, param, context, expected):
        assert param.render(context) == expected

    @pytest.mark.parametrize(
        "bit,expected",
        [
            ("class='btn'", Param(name="class", value="btn")),
            ('class="btn"', Param(name="class", value="btn")),
            ("disabled", Param(name="disabled", value=True)),
        ],
    )
    def test_from_bit(self, bit, expected):
        assert Param.from_bit(bit) == expected


class TestParams:
    @pytest.mark.parametrize(
        "params,nodelist,context,expected_props,expected_attrs",
        [
            (
                Params(attrs=[Param(name="class", value=None)]),
                [PropNode(name="class", default="btn", attrs=[])],
                {},
                {"class": "btn"},
                [],
            ),
            (
                Params(attrs=[Param(name="class", value="btn")]),
                [PropNode(name="class", default=None, attrs=[])],
                {},
                {"class": "btn"},
                [],
            ),
            (
                Params(attrs=[Param(name="class", value="btn")]),
                None,
                {},
                None,
                [Param(name="class", value="btn")],
            ),
        ],
    )
    def test_render_props(
        self, params, nodelist, context, expected_props, expected_attrs
    ):
        assert params.render_props(nodelist, context) == expected_props
        assert params.attrs == expected_attrs

    @pytest.mark.parametrize(
        "params,context,expected",
        [
            (Params(attrs=[Param(name="class", value="btn")]), {}, 'class="btn"'),
            (
                Params(attrs=[Param(name="class", value="btn")]),
                {"btn": "blue"},
                'class="blue"',
            ),
            (Params(attrs=[Param(name="disabled", value=True)]), {}, "disabled"),
            (
                Params(
                    attrs=[
                        Param(name="class", value="btn"),
                        Param(name="disabled", value=True),
                    ]
                ),
                {},
                'class="btn" disabled',
            ),
        ],
    )
    def test_render_attrs(self, params, context, expected):
        assert params.render_attrs(context) == expected

    @pytest.mark.parametrize(
        "bits,expected",
        [
            (["class='btn'"], Params(attrs=[Param(name="class", value="btn")])),
            (['class="btn"'], Params(attrs=[Param(name="class", value="btn")])),
            (["disabled"], Params(attrs=[Param(name="disabled", value=True)])),
        ],
    )
    def test_from_bits(self, bits, expected):
        assert Params.from_bits(bits) == expected



from __future__ import annotations

import pytest
from django.conf import settings
from django.test import override_settings

from django_bird.conf import AppSettings
from django_bird.conf import TemplateConfigurator
from django_bird.conf import app_settings


def test_app_settings():
    assert app_settings.ENABLE_AUTO_CONFIG is True


@override_settings(
    DJANGO_BIRD={
        "ENABLE_AUTO_CONFIG": False,
    }
)
def test_autoconfigure_disabled():
    app_settings = AppSettings()
    template_options = settings.TEMPLATES[0]["OPTIONS"]

    assert app_settings.ENABLE_AUTO_CONFIG is False
    assert "django_bird.templatetags.django_bird" not in template_options["loaders"]
    assert "django_bird.loader.BirdLoader" not in template_options["builtins"]


class TestTemplateConfigurator:
    DJANGO_BIRD_BUILTINS = "django_bird.templatetags.django_bird"
    DJANGO_BIRD_LOADER = "django_bird.loader.BirdLoader"

    @pytest.fixture
    def configurator(self):
        return TemplateConfigurator(app_settings)

    @pytest.fixture(autouse=True)
    def reset_settings(self):
        template_options = settings.TEMPLATES[0]["OPTIONS"]

        assert self.DJANGO_BIRD_LOADER in template_options["loaders"]
        assert self.DJANGO_BIRD_BUILTINS in template_options["builtins"]

        with override_settings(
            TEMPLATES=[
                settings.TEMPLATES[0]
                | {
                    **settings.TEMPLATES[0],
                    "OPTIONS": {
                        "loaders": [
                            loader
                            for loader in template_options["loaders"]
                            if loader != self.DJANGO_BIRD_LOADER
                        ],
                        "builtins": [
                            builtin
                            for builtin in template_options["builtins"]
                            if builtin != self.DJANGO_BIRD_BUILTINS
                        ],
                    },
                }
            ]
        ):
            options = settings.TEMPLATES[0]["OPTIONS"]

            assert self.DJANGO_BIRD_LOADER not in options["loaders"]
            assert self.DJANGO_BIRD_BUILTINS not in options["builtins"]

            yield

    def test_autoconfigure(self, configurator):
        template_options = settings.TEMPLATES[0]["OPTIONS"]

        configurator.autoconfigure()

        assert self.DJANGO_BIRD_LOADER in template_options["loaders"]
        assert self.DJANGO_BIRD_BUILTINS in template_options["builtins"]

    def test_configure_loaders(self, configurator):
        template_options = settings.TEMPLATES[0]["OPTIONS"]

        configurator.configure_loaders(template_options)

        assert self.DJANGO_BIRD_LOADER in template_options["loaders"]

    def test_configure_builtins(self, configurator):
        template_options = settings.TEMPLATES[0]["OPTIONS"]

        configurator.configure_builtins(template_options)

        assert self.DJANGO_BIRD_BUILTINS in template_options["builtins"]

    def test_configured(self, configurator):
        assert configurator._configured is False

        configurator.autoconfigure()

        assert configurator._configured is True

    @pytest.mark.parametrize(
        "init_options,expected",
        [
            (
                {
                    "builtins": [
                        "django.template.defaulttags",
                    ],
                    "loaders": [
                        "django.template.loaders.filesystem.Loader",
                        "django.template.loaders.app_directories.Loader",
                    ],
                },
                {
                    "builtins": [
                        "django.template.defaulttags",
                        "django_bird.templatetags.django_bird",
                    ],
                    "loaders": [
                        "django_bird.loader.BirdLoader",
                        "django.template.loaders.filesystem.Loader",
                        "django.template.loaders.app_directories.Loader",
                    ],
                },
            ),
            (
                {
                    "builtins": [
                        "django.template.defaulttags",
                    ],
                    "loaders": [
                        (
                            "django.template.loaders.cached.Loader",
                            [
                                "django.template.loaders.filesystem.Loader",
                                "django.template.loaders.app_directories.Loader",
                            ],
                        ),
                    ],
                },
                {
                    "builtins": [
                        "django.template.defaulttags",
                        "django_bird.templatetags.django_bird",
                    ],
                    "loaders": [
                        (
                            "django.template.loaders.cached.Loader",
                            [
                                "django_bird.loader.BirdLoader",
                                "django.template.loaders.filesystem.Loader",
                                "django.template.loaders.app_directories.Loader",
                            ],
                        ),
                    ],
                },
            ),
            (
                {
                    "builtins": [
                        "django.template.defaulttags",
                    ],
                    "loaders": [
                        (
                            "template_partials.loader.Loader",
                            [
                                (
                                    "django.template.loaders.cached.Loader",
                                    [
                                        "django.template.loaders.filesystem.Loader",
                                        "django.template.loaders.app_directories.Loader",
                                    ],
                                ),
                            ],
                        )
                    ],
                },
                {
                    "builtins": [
                        "django.template.defaulttags",
                        "django_bird.templatetags.django_bird",
                    ],
                    "loaders": [
                        (
                            "template_partials.loader.Loader",
                            [
                                (
                                    "django.template.loaders.cached.Loader",
                                    [
                                        "django_bird.loader.BirdLoader",
                                        "django.template.loaders.filesystem.Loader",
                                        "django.template.loaders.app_directories.Loader",
                                    ],
                                ),
                            ],
                        )
                    ],
                },
            ),
            (
                {
                    "builtins": [
                        "django.template.defaulttags",
                        "django_cotton.templatetags.cotton",
                    ],
                    "loaders": [
                        (
                            "django.template.loaders.cached.Loader",
                            [
                                "django_cotton.cotton_loader.Loader",
                                "django.template.loaders.filesystem.Loader",
                                "django.template.loaders.app_directories.Loader",
                            ],
                        ),
                    ],
                },
                {
                    "builtins": [
                        "django.template.defaulttags",
                        "django_cotton.templatetags.cotton",
                        "django_bird.templatetags.django_bird",
                    ],
                    "loaders": [
                        (
                            "django.template.loaders.cached.Loader",
                            [
                                "django_bird.loader.BirdLoader",
                                "django_cotton.cotton_loader.Loader",
                                "django.template.loaders.filesystem.Loader",
                                "django.template.loaders.app_directories.Loader",
                            ],
                        ),
                    ],
                },
            ),
            (
                {
                    "builtins": [
                        "django.template.defaulttags",
                        "django_cotton.templatetags.cotton",
                    ],
                    "loaders": [
                        (
                            "template_partials.loader.Loader",
                            [
                                (
                                    "django.template.loaders.cached.Loader",
                                    [
                                        "django_cotton.cotton_loader.Loader",
                                        "django.template.loaders.filesystem.Loader",
                                        "django.template.loaders.app_directories.Loader",
                                    ],
                                ),
                            ],
                        )
                    ],
                },
                {
                    "builtins": [
                        "django.template.defaulttags",
                        "django_cotton.templatetags.cotton",
                        "django_bird.templatetags.django_bird",
                    ],
                    "loaders": [
                        (
                            "template_partials.loader.Loader",
                            [
                                (
                                    "django.template.loaders.cached.Loader",
                                    [
                                        "django_bird.loader.BirdLoader",
                                        "django_cotton.cotton_loader.Loader",
                                        "django.template.loaders.filesystem.Loader",
                                        "django.template.loaders.app_directories.Loader",
                                    ],
                                ),
                            ],
                        )
                    ],
                },
            ),
        ],
    )
    def test_template_settings(self, init_options, expected, configurator):
        with override_settings(
            TEMPLATES=[
                settings.TEMPLATES[0]
                | {
                    **settings.TEMPLATES[0],
                    "OPTIONS": init_options,
                }
            ]
        ):
            template_options = settings.TEMPLATES[0]["OPTIONS"]

            assert template_options == init_options
            assert self.DJANGO_BIRD_LOADER not in template_options["loaders"]
            assert self.DJANGO_BIRD_BUILTINS not in template_options["builtins"]

            configurator.autoconfigure()

            assert template_options == expected



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path

from django.templatetags.static import static
from django.utils.html import format_html
from django.utils.html import format_html_join
from django.utils.safestring import SafeString

from ._typing import override


class AssetType(IntEnum):
    CSS = 1
    JS = 2


@dataclass(frozen=True)
class Asset:
    path: Path
    type: AssetType

    @override
    def __hash__(self) -> int:
        return hash((str(self.path), self.type))

    def exists(self) -> bool:
        return self.path.exists()

    @classmethod
    def from_path(cls, path: Path) -> Asset:
        match path.suffix.lower():
            case ".css":
                asset_type = AssetType.CSS
            case ".js":
                asset_type = AssetType.JS
            case _:
                raise ValueError(f"Unknown asset type for path: {path}")
        return cls(path=path, type=asset_type)


class Registry:
    def __init__(self) -> None:
        self.assets: set[Asset] = set()

    def register(self, asset: Asset | Path) -> None:
        if isinstance(asset, Path):
            asset = Asset.from_path(asset)

        if not asset.exists():
            raise FileNotFoundError(f"Asset file not found: {asset.path}")

        self.assets.add(asset)

    def clear(self) -> None:
        self.assets.clear()

    def get_assets(self, asset_type: AssetType) -> list[Asset]:
        assets = [asset for asset in self.assets if asset.type == asset_type]
        return self.sort_assets(assets)

    def sort_assets(
        self,
        assets: list[Asset],
        *,
        key: Callable[[Asset], str] = lambda a: str(a.path),
    ) -> list[Asset]:
        return sorted(assets, key=key)

    def get_format_string(self, asset_type: AssetType) -> str:
        match asset_type:
            case AssetType.CSS:
                return '<link rel="stylesheet" href="{}" type="text/css">'
            case AssetType.JS:
                return '<script src="{}" type="text/javascript">'

    def render(self, asset_type: AssetType) -> SafeString:
        assets = self.get_assets(asset_type)

        if not assets:
            return format_html("")

        return format_html_join(
            "\n",
            self.get_format_string(asset_type),
            ((static(str(asset.path)),) for asset in assets),
        )


registry = Registry()



from __future__ import annotations

import pytest
from django.template.loader import get_template


def test_create_bird_template(create_bird_template):
    name = "foo"
    content = "<div>bar</div>"

    create_bird_template(name, content)

    template = get_template(f"bird/{name}.html")

    assert template
    assert template.render({}) == content


@pytest.mark.parametrize(
    "contents,expected",
    [
        ("<button> Click me </button>", "<button>Click me</button>"),
        ("<button> Click me</button>", "<button>Click me</button>"),
        ("<button>Click me </button>", "<button>Click me</button>"),
        ("<button >Click me</button>", "<button>Click me</button>"),
        (
            "<button class='btn'> Click me </button>",
            "<button class='btn'>Click me</button>",
        ),
        ("\n<button>\n  Click me\n</button>\n", "<button>Click me</button>"),
    ],
)
def test_normalize_whitespace(contents, expected, normalize_whitespace):
    assert normalize_whitespace(contents) == expected
